import 'dart:io';

import 'package:path/path.dart';
import 'package:protobuf3_parser/protobuf3_parser.dart';
import 'package:protobuf3_parser/src/asset_manager/asset_reader.dart';
import 'package:protobuf3_parser/src/asset_manager/asset_writer.dart';
import 'package:protobuf3_parser/src/asset_manager/asset_id.dart';
import 'package:protobuf3_parser/src/asset_manager/directory_asset_reader.dart';
import 'package:protobuf3_parser/src/asset_manager/directory_asset_writer.dart';
import 'package:protobuf3_parser/src/extensions.dart';
import 'package:protobuf3_parser/src/gen/doc_gen/doc_generator.dart';

extension on Element {
  String get path {
    String path = "${runtimeType}_$name";
    Element? parent = parentElement;
    while (parent != null) {
      path = "${parent.runtimeType}_${parent.name}__$path";
      parent = parent.parentElement;
    }
    return path;
  }
}

class MarkdownDocGenerator extends DocGenerator {
  MarkdownDocGenerator(
      AssetId assetId, AssetReader assetReader, AssetWriter assetWriter)
      : super(assetId, assetReader, assetWriter);

  MarkdownDocGenerator.fromFile(File file)
      : super(AssetId(basename(file.path)), DirectoryAssetReader(file.parent),
            DirectoryAssetWriter(join(file.parent.path, "doc").dir));

  Stream<List<int>> _writeComments(List<Comment> comments,
      [String? prefix]) async* {
    prefix ??= '> ';

    for (int i = 0; i < comments.length; i++) {
      final comment = comments[i];
      final isLast = i == comments.length - 1;

      for (final line in comment.formated.split("\n")) {
        if (line.trim().isEmpty) continue;
        yield l("$prefix$line");
      }

      if (!isLast) {
        if (comments[i + 1].formated.isNotEmpty) {
          // yield l();
        }
      }
    }
    // yield l("```");
  }

  Stream<List<int>> _writeElementComments(Element el, [String? prefix]) async* {
    if (el.hasComments) {
      yield* _writeComments(el.comments, prefix);
    }
  }

  List<int> l([String? text]) => w("${text ?? ''}\n");
  List<int> w(String text) => text.codeUnits;
  String _linkFile(String path, String text) => "<a href='$path'>$text</a>";
  String _linkMd(String path, String text) => _linkFile("$path.md", text);
  String _b(String text) => "<b>$text</b>";
  String _i(String text) => "<i>$text</i>";
  List<int> _hr() => [...l("<hr/>"), ...l()];
  String span(String css, String text) => "<span style=\"$css\">$text</span>";

  @override
  Future<void> generate() async {
    await for (final el in parser.allElementsInExistence) {
      if (el is Message) {
        await _generateMessage(el);
      } else if (el is Enum) {
        await _generateEnum(el);
      } else if (el is Service) {
        await _generateService(el);
      }
    }

    for (final file in [...await parser.allFilesInExistence.toList(), parser]) {
      await _generateFile(file.assetId);
    }
    await _generateIndex();
  }

  Stream<List<int>> _writeTree(Element el) async* {
    var path =
        "${_linkMd("index", "index")} < ${_linkMd("File_${el.parser.assetId.path}", basename(el.parser.assetId.path))}";

    String? p;

    // parent
    Element? parent = el.parentElement;
    while (parent != null) {
      p = "${_linkMd(parent.path, _b(parent.name))} ${p == null ? '' : '< $p'}";
      parent = parent.parentElement;
    }

    yield w("$path ${p == null ? '' : '< $p'}");
  }

  /// =========== MESSAGE

  Future _generateMessage(Message message) async {
    final id = AssetId(join("${message.path}.md"));
    await assetWriter.writeStream(id, __generateMessage(message));
  }

  Stream<List<int>> __generateMessage(Message message) async* {
    yield l("<!-- Auto generated. Do not modify. -->");
    yield* _writeTree(message);
    yield l(' < `message`');

    yield l("# ${message.name}");
    yield* _writeElementComments(message, '');

    if (message.fields.isNotEmpty) {
      yield _hr();

      // fields
      for (final field in message.fields) {
        final fieldType = await field.type;

        yield* _writeElementComments(field);

        yield w("* ${field.isRepeated ? 'repeated ' : ''}");
        if (fieldType.isCore) {
          yield l("${_i(fieldType.name)} ${field.name} = ${field.id};");
        } else {
          if (fieldType.isEnum) {
            yield l(
                "${_linkMd(fieldType.path, _b(fieldType.name))} ${field.name} = ${field.id};");
          } else {
            yield l(
                "${_linkMd(fieldType.path, _b(fieldType.name))} ${field.name} = ${field.id};");
          }
        }

        yield l();
      }
    }

    if (message.subMessages.isNotEmpty) {
      yield _hr();
      yield l();

      for (final subMessage in message.subMessages) {
        yield l(
            "* message ${_linkMd(subMessage.path, _b(subMessage.name))} { ... }");
      }
    }

    if (message.subEnums.isNotEmpty) {
      yield _hr();

      for (final subEnum in message.subEnums) {
        yield l("* enum ${_linkMd(subEnum.path, _b(subEnum.name))} { ... }");
      }
    }
  }

  /// =========== ENUM

  Future _generateEnum(Enum enum_) async {
    final id = AssetId("${enum_.path}.md");
    await assetWriter.writeStream(id, __generateEnum(enum_));
  }

  Stream<List<int>> __generateEnum(Enum enum_) async* {
    yield l("<!-- Auto generated. Do not modify. -->");
    yield* _writeTree(enum_);
    yield l(' < `enum`');

    yield l("# ${enum_.name}");
    yield* _writeElementComments(enum_, '');

    if (enum_.fields.isNotEmpty) {
      yield _hr();

      for (final field in enum_.fields) {
        yield* _writeElementComments(field);
        yield l("* ${field.name} = ${field.id};");

        yield l();
      }
    }
  }

  /// =========== SERVICE

  Future _generateService(Service service) async {
    final id = AssetId("${service.path}.md");
    await assetWriter.writeStream(id, __generateService(service));
  }

  Stream<List<int>> __generateService(Service service) async* {
    yield l("<!-- Auto generated. Do not modify. -->");
    yield* _writeTree(service);
    yield l(' < `service`');

    yield l("# ${service.name}");
    yield* _writeElementComments(service, '');
    yield _hr();

    for (final method in service.methods) {
      final methodReq = await method.request;
      final methodResp = await method.response;

      yield* _writeElementComments(method);
      yield l(
          "* ${method.name} ( ${method.isStreamRequest ? "stream " : ""} ${_linkMd(methodReq.path, _b(methodReq.name))} ) returns ( ${method.isStreamResponse ? "stream " : ""} ${_linkMd(methodResp.path, _b(methodResp.name))} );");

      yield l();
    }
  }

  /// =========== FILE

  Future _generateFile(AssetId assetId) async {
    final id = AssetId("File_${split(assetId.path).join("_")}.md");
    await assetWriter.writeStream(id, __generateFile(assetId));
  }

  Stream<List<int>> __generateFile(AssetId assetId) async* {
    yield l("<!-- Auto generated. Do not modify. -->");

    yield l("${_linkMd("index", "index")} < `file`");
    yield l("# ${assetId.path}");

    final comments = await parser.getFile(assetId).fileComments;
    if (comments.isNotEmpty) {
      yield* _writeComments(comments, '');
    }
    yield _hr();

    final file = parser.getFile(assetId);

    await for (final message in file.messages) {
      yield* _writeElementComments(message);
      yield l("* message ${_linkMd(message.path, message.name)} { ... }");
    }

    await for (final enum_ in file.enums) {
      yield* _writeElementComments(enum_);
      yield l("* enum ${_linkMd(enum_.path, enum_.name)} { ... }");
    }

    await for (final service in file.services) {
      yield* _writeElementComments(service);
      yield l("* service ${_linkMd(service.path, service.name)} { ... }");
    }
  }

  /// =========== INDEX

  Future _generateIndex() async {
    await assetWriter.writeStream(AssetId("index.md"), __generateIndex());
  }

  Stream<List<int>> __generateIndex() async* {
    yield l("<!-- Auto generated. Do not modify. -->");
    yield l("# Index");

    for (final file in [
      parser,
      ...(await parser.allFilesInExistence.toList())
    ]) {
      yield* _writeComments(await file.fileComments);
      yield l(
          "* ${_linkMd('File_${file.assetId.path}', _b(basename(file.assetId.path)))} { ... }");
    }
  }
}
