import 'package:protobuf3_parser/protobuf3_parser.dart';

final defaultSrcGen = DefaultSourceGenerator();

class DefaultSourceGenerator extends SourceGenerator {
  void _writeComments(Element element, StringBuffer x, String indent) {
    for (final comment in element.comments) {
      for (final line in comment.original.split("\n")) {
        x.writeln("$indent$line");
      }
    }
  }

  // ======================== MESSAGE ==============================

  @override
  Future<String> forMessage(Message message, [int indentLevel = 0]) async {
    final indent = "\t" * indentLevel;

    final x = StringBuffer();

    // comments
    _writeComments(message, x, indent);

    // message NAME {
    x.writeln("${indent}message ${message.name} {");

    // fields
    for (final field in message.fields) {
      x.write(await forMessageField(field, indentLevel + 1));
    }

    // sub enums.
    for (final subEnum in message.subEnums) {
      x
        ..writeln()
        ..write(forEnum(subEnum, indentLevel + 1));
    }

    // sub messages.
    for (final subMessage in message.subMessages) {
      x
        ..writeln()
        ..write(await forMessage(subMessage, indentLevel + 1));
    }

    // }
    x.writeln("$indent}");

    return x.toString();
  }

  @override
  Future<String> forMessageField(MessageField field,
      [int indentLevel = 0]) async {
    final indent = "\t" * indentLevel;

    final x = StringBuffer();
    // comments
    _writeComments(field, x, indent);

    // field
    x.writeln("$indent${(await field.type).name} ${field.name} = ${field.id};");
    return x.toString();
  }

  /// ============================================================
  /// ======================== ENUM ==============================
  /// ============================================================

  @override
  String forEnum(Enum enum_, [int indentLevel = 0]) {
    final indent = "\t" * indentLevel;

    final x = StringBuffer();

    // comments
    _writeComments(enum_, x, indent);

    // enum NAME {
    x.writeln("${indent}enum ${enum_.name} {");

    // fields
    for (final field in enum_.fields) {
      x.write(forEnumField(field, indentLevel + 1));
    }

    x.writeln("$indent}");

    return x.toString();
  }

  @override
  String forEnumField(EnumField field, [int indentLevel = 0]) {
    final indent = "\t" * indentLevel;

    final x = StringBuffer();
    // field comments
    _writeComments(field, x, indent);

    // fields
    x.writeln("$indent${field.name} = ${field.id};");
    return x.toString();
  }

  /// ============================================================
  /// ===================== SERVICE ==============================
  /// ============================================================

  @override
  Future<String> forService(Service service, [int indentLevel = 0]) async {
    final indent = "\t" * indentLevel;

    final x = StringBuffer();

    // service NAME {
    x.writeln("${indent}service ${service.name} {");

    // methods
    for (final method in service.methods) {
      x.write(await forServiceMethod(method, indentLevel + 1));
    }

    // }
    x.writeln("$indent}");
    return x.toString();
  }

  @override
  Future<String> forServiceMethod(ServiceMethod method,
      [int indentLevel = 0]) async {
    final indent = "\t" * indentLevel;

    final x = StringBuffer();

    // comments
    _writeComments(method, x, indent);

    // grpc
    x.write("${indent}rpc ${method.name} (");
    if (method.isStreamRequest) {
      x.write("stream ");
    }
    x.write("${(await method.request).name}) returns (");
    if (method.isStreamResponse) {
      x.write("stream ");
    }
    x.writeln("${(await method.response).name});");

    return x.toString();
  }
}
