import 'dart:io';

import 'package:antlr4/antlr4.dart';
import 'package:path/path.dart';
// ignore: library_prefixes
import 'package:protobuf3_parser/__generated__/antlr/Protobuf3Parser.dart' as $;
// ignore: library_prefixes
import 'package:protobuf3_parser/__generated__/antlr/Protobuf3Lexer.dart' as $;
import 'package:protobuf3_parser/protobuf3_parser.dart';
import 'package:protobuf3_parser/src/asset_manager/directory_asset_reader.dart';
import '../asset_manager/asset_id.dart';
import '../asset_manager/asset_reader.dart';
import '../asset_manager/bytes_asset_reader.dart';

class Parser {
  AssetId assetId;
  AssetReader assetReader;
  Parser(this.assetId, this.assetReader);

  Parser.fromString(String string)
      : assetId = AssetId(""),
        assetReader = BytesAssetReader(string.codeUnits);

  Parser.fromBytes(List<int> bytes)
      : assetId = AssetId(""),
        assetReader = BytesAssetReader(bytes);
  Parser.fromFile(File file)
      : assetId = AssetId(basename(file.path)),
        assetReader = DirectoryAssetReader(file.parent);

  // ========================================
  // API:
  // ========================================

  Parser getFile(AssetId id) {
    return Parser(id, assetReader);
  }

  /// Get all the imported files.
  Stream<Parser> get files async* {
    // cancel, if the assetReader is bytes.
    if (assetReader is BytesAssetReader) return;

    final tree = await _tree;
    for (final importStatement in tree.importStatements()) {
      final path = importStatement.strLit()?.text;

      if (path != null) {
        yield Parser(AssetId(path.substring(1, path.length - 1)), assetReader);
      }
    }
  }

  /// Get all the toplevel [Message]s.
  Stream<Message> get messages async* {
    await for (final messageDef in _getMessageDefs()) {
      yield convertToMessage(messageDef);
    }
  }

  /// Get all the toplevel [Enum]s.
  Stream<Enum> get enums async* {
    await for (final enumDef in _getEnumDefs()) {
      yield convertToEnum(enumDef);
    }
  }

  /// Get all the toplevel [Service]s.
  Stream<Service> get services async* {
    await for (final messageDef in _getServiceDefs()) {
      yield await convertToService(messageDef);
    }
  }

  /// Get all the toplevel [Element]s.
  Stream<Element> get elements async* {
    yield* messages;
    yield* enums;
    yield* services;
  }

  // ========================================
  // API: MESSAGE HELPERS
  // ========================================

  /// Get all the [Message]s in this file.
  Stream<Message> get allMessagesInThisFile async* {
    await for (final message in messages) {
      yield message;
      yield* _getMessageMessages(message, true);
    }
  }

  /// Get all the [Message]s in this file and other imported files.
  Stream<Message> get allMessagesInExistence async* {
    yield* allMessagesInThisFile;

    await for (final file in files) {
      yield* file.allMessagesInExistence;
    }
  }

  Stream<Message> _getMessageMessages(Message message,
      [bool sub = true]) async* {
    for (final subMessage in message.subMessages) {
      yield subMessage;

      if (sub) {
        yield* _getMessageMessages(subMessage, true);
      }
    }
  }

  Stream<Enum> _getMessageEnums(Message message, [bool sub = true]) async* {
    for (final subEnum in message.subEnums) {
      yield subEnum;
    }

    if (sub) {
      for (final subMessage in message.subMessages) {
        yield* _getMessageEnums(subMessage, true);
      }
    }
  }

  // ========================================
  // API: Element HELPERS
  // ========================================

  /// Get all the [Element]s in this file.
  Stream<Element> get allElementsInThisFile async* {
    yield* allMessagesInThisFile;
    yield* allEnumInThisFile;
    yield* allServiceInThisFile;
  }

  /// Get all the [Element]s in this file and imported files.
  Stream<Element> get allElementsInExistence async* {
    yield* allMessagesInExistence;
    yield* allEnumsInExistence;
    yield* allServicesInExistence;

    await for (final file in files) {
      yield* file.allElementsInExistence;
    }
  }

  // ========================================
  // API: SERVICE HELPERS
  // ========================================

  /// Get all the [Service]s in this file.
  Stream<Service> get allServiceInThisFile => services;

  /// Get all the [Service]s in this file and other imported files.
  Stream<Service> get allServicesInExistence async* {
    yield* services;

    await for (final file in files) {
      yield* file.allServicesInExistence;
    }
  }

  // ========================================
  // API: ENUM HELPERS
  // ========================================

  /// Get all the [Enum]s in this file.
  Stream<Enum> get allEnumInThisFile async* {
    yield* enums;

    await for (final message in messages) {
      yield* _getMessageEnums(message, true);
    }
  }

  /// Get all the [Enum]s in this file and other imported files.
  Stream<Enum> get allEnumsInExistence async* {
    yield* allEnumInThisFile;

    await for (final file in files) {
      yield* file.allEnumsInExistence;
    }
  }

  // ========================================
  // API: Parser HELPERS
  // ========================================

  /// Get all the [Service]s in this file.
  Stream<Parser> get allFilesInThisFile => files;

  /// Get all the [Service]s in this file and other imported files.
  Stream<Parser> get allFilesInExistence async* {
    yield* allFilesInThisFile;

    await for (final file in files) {
      yield* file.allFilesInExistence;
    }
  }
  // ========================================
  // API: GENERATORS
  // ========================================

  /// Generate source code stream.
  Stream<List<int>> toSourceStream() async* {
    await for (final element in elements) {
      yield (await element.toSource()).codeUnits;
    }
  }

  /// Generate source code.
  Future<String> toSource() async {
    final x = StringBuffer();
    await for (final element in elements) {
      x.write(await element.toSource());
    }
    return x.toString();
  }

  /// Generate source code stream.
  Stream<List<int>> toMiniSourceStream() async* {
    await for (final element in elements) {
      yield (await element.toMiniSource()).codeUnits;
    }
  }

  /// Generate minified source code.
  Future<String> toMiniSource() async {
    final x = StringBuffer();
    await for (final element in elements) {
      x.write(await element.toMiniSource());
    }
    return x.toString();
  }

  // ========================================
  // UTILS
  // ========================================

  /// Get the main proto ctx.
  Future<$.ProtoContext> get _tree async {
    final input = await InputStream.fromStream(assetReader.read(assetId));
    final lexer = $.Protobuf3Lexer(input);
    final tokens = CommonTokenStream(lexer);
    final parser = $.Protobuf3Parser(tokens);
    return parser.proto();
  }

  /// Get all the toplevel [$.MessageDefContext]s.
  Stream<$.MessageDefContext> _getMessageDefs() async* {
    final tree = await _tree;
    for (final tree in tree.topLevelDefs()) {
      final messageDef = tree.messageDef();
      if (messageDef != null) {
        yield messageDef;
      }
    }
  }

  /// Get all the toplevel [$.EnumDefContext]s.
  Stream<$.EnumDefContext> _getEnumDefs() async* {
    final tree = await _tree;
    for (final tree in tree.topLevelDefs()) {
      final enumDef = tree.enumDef();
      if (enumDef != null) {
        yield enumDef;
      }
    }
  }

  /// Get all the toplevel [$.ServiceDefContext]s.
  Stream<$.ServiceDefContext> _getServiceDefs() async* {
    final tree = await _tree;
    for (final tree in tree.topLevelDefs()) {
      final serviceDef = tree.serviceDef();
      if (serviceDef != null) {
        yield serviceDef;
      }
    }
  }

  // ========================================
  // CONVERTS
  // ========================================

  /// Convert [$.MessageDefContext] to [Message].
  Message convertToMessage($.MessageDefContext ctx) => Message(ctx, this);

  /// Convert [$.EnumDefContext] to [Enum].
  Enum convertToEnum($.EnumDefContext ctx) => Enum(ctx, this);

  /// Convert [$.ServiceDefContext] to [Service].
  Service convertToService($.ServiceDefContext ctx) => Service(ctx, this);

  /// Convert [$.CommentContext]s to [Comment]s.
  List<Comment> convertToComments(List<$.CommentContext> ctxs) => ctxs
      .map((x) => convertToComment(x))
      .where((x) => x != null)
      .map((e) => e!)
      .toList();

  /// Convert [$.CommentContext] to [Comment].
  Comment? convertToComment($.CommentContext ctx) {
    final line = ctx.LINE_COMMENT()?.text;
    final multi = ctx.MULTI_COMMENT()?.text;
    final lineDoc = ctx.DOC_LINE_COMMENT()?.text;
    final multiDoc = ctx.DOC_MULTI_COMMENT()?.text;

    if (line != null) {
      return Comment(CommentType.line, line);
    } else if (lineDoc != null) {
      return Comment(CommentType.lineDoc, lineDoc);
    } else if (multi != null) {
      return Comment(CommentType.multiDoc, multi);
    } else if (multiDoc != null) {
      return Comment(CommentType.multiDoc, multiDoc);
    } else {
      return null;
    }
  }

  Future<List<Comment>> get fileComments async =>
      convertToComments((await _tree).syntax()?.comments() ?? []);
}
